@@ -6,6 +6,7 @@ import time                  # built-in library
 import warnings              # built-in library
 from typing import Tuple     # built-in library
 from typing import Sized     # built-in library
+from typing import Iterable  # built-in library
 from typing import Optional  # built-in library
 from typing import Callable  # built-in library
 import numpy as np           # pip install numpy
@@ -104,6 +105,27 @@ class Algorithms:
             self._vprint(f"{_elapsed(t0)} - resizing [Lanczos] from {orgw} x {orgh} to {dstw} x {dsth}")
         return frame
 
+    def subtract(self, frame: np.ndarray, blacklevels: Iterable) -> np.ndarray:
+        """
+        Subtract per-channel black levels from the given frame, but do not linearize.
+        For demosaiced RGB frames, blacklevels must contain three values; for raw Bayer
+        frames, four values are required. The caller is responsible for making sure
+        that the per-channel levels are in the same Bayer order as the frame itself.
+        """
+        t0 = time.time()
+        levels = np.array(blacklevels)
+        if frame.ndim == 3:  # already demosaiced
+            assert levels.size == 3, levels
+            frame = np.maximum(frame, levels)
+            frame = frame - levels
+        else:  # raw Bayer
+            assert levels.size == 4, levels
+            rggb = list(self.bayer_split(frame))
+            rggb = [np.maximum(ch, level) - level for ch, level in zip(rggb, levels)]
+            frame = self.bayer_combine(*rggb)
+        self._vprint(f"{_elapsed(t0)} - subtracting black levels {levels}: range = {self._minmax(frame)}")
+        return frame
+
     def linearize(self,
                   frame: np.ndarray,
                   blacklevel: float = None,
@@ -118,6 +140,7 @@ class Algorithms:
         This algorithm is format-agnostic, although it's typically applied on raw
         sensor data.
         """
+        minmax = self._minmax(frame)
         if blacklevel is None:
             t0 = time.time()
             percentile = num_clipped / frame.size * 100.0
@@ -128,12 +151,12 @@ class Algorithms:
             percentile = (1.0 - num_clipped / frame.size) * 100.0
             whitelevel = np.percentile(frame, percentile)
             self._vprint(f"{_elapsed(t0)} - estimating white level: {percentile:5.2f}th percentile = {whitelevel:.2f}")
-        if (blacklevel, whitelevel) != (0, 1):
-            frame = self.clip(frame, blacklevel, whitelevel)
-            t0 = time.time()
-            frame -= blacklevel
-            frame /= whitelevel - blacklevel
-            self._vprint(f"{_elapsed(t0)} - linearizing from [{blacklevel:.2f}, {whitelevel:.2f}] to [0, 1]")
+        assert whitelevel > blacklevel, f"{whitelevel} is not greater than {blacklevel}"
+        frame = np.clip(frame, blacklevel, whitelevel)
+        t0 = time.time()
+        frame -= blacklevel
+        frame /= whitelevel - blacklevel
+        self._vprint(f"{_elapsed(t0)} - linearizing from {minmax} to [{blacklevel:.2f}, {whitelevel:.2f}] to [0, 1]")
         return frame
 
     def demosaic(self,
@@ -227,16 +250,18 @@ class Algorithms:
                 self._vprint(f"{_elapsed(t0)} - applying WB gains {gains}: range = {self._minmax(frame)}")
         return frame
 
-    def ccm(self, frame: np.ndarray, matrix: np.ndarray) -> np.ndarray:
+    def ccm(self, frame: np.ndarray, matrix: np.ndarray, clip=True) -> np.ndarray:
         """
-        Apply the given color correction matrix on the given RGB frame. Input
-        colors are clipped to [0, 1] to avoid "pink sky" artifacts caused by the
-        combination of clipped highlights and less-than-1.0 coefficients in the
-        CCM. In other words, no attempt is made at gamut mapping or highlight
-        recovery. If matrix is None, the frame is returned untouched.
+        Apply the given color correction matrix on the given RGB frame. If the
+        'clip' flag is True, input colors are clipped to [0, 1] to avoid "pink
+        sky" artifacts caused by the combination of clipped highlights and
+        less-than-1.0 coefficients in the CCM. No attempt is made at gamut
+        mapping or highlight recovery. If matrix is None, the frame is returned
+        untouched.
         """
         if matrix is not None:
-            frame = self.clip(frame, 0, 1)
+            if clip:
+                frame = self.clip(frame, 0, 1)
             t0 = time.time()
             frame = frame @ matrix.T
             with np.printoptions(formatter={'float': lambda x: f"{x:.2f}"}):
@@ -314,9 +339,10 @@ class Algorithms:
         If mode is None, the frame is returned untouched.
         """
         if mode is not None:
-            realmode = mode
-            frame = self.clip(frame, 0, 1)  # can't handle values outside of [0, 1]
             t0 = time.time()
+            realmode = mode
+            input_range = self._minmax(frame)
+            frame = np.clip(frame, 0, 1)  # can't handle values outside of [0, 1]
             if realmode in ["sRGB", "rec709"]:
                 bpp = 14
                 maxval = 2 ** bpp - 1
@@ -337,15 +363,15 @@ class Algorithms:
             if mode == "LUT":
                 assert lut is not None
                 maxval = len(lut) - 1
-                frame = self.quantize(frame, maxval)  # [0, 1] ==> [0, maxval]
-                frame = lut[frame]                    # [0, maxval] ==> [0, maxval]
-                frame = frame / maxval                # [0, maxval] ==> [0, 1]
-            self._vprint(f"{_elapsed(t0)} - applying gamma curve [{realmode}]: range = {self._minmax(frame)}")
+                frame = self._quantize(frame, maxval)  # [0, 1] ==> [0, maxval]
+                frame = lut[frame]                     # [0, maxval] ==> [0, maxval]
+                frame = frame / maxval                 # [0, maxval] ==> [0, 1]
+            self._vprint(f"{_elapsed(t0)} - applying gamma curve [{realmode}]: input range = {input_range}")
         return frame
 
     def quantize(self,
                  frame: np.ndarray,
-                 maxval: int,
+                 maxval: int = 65535,
                  dtype: np.dtype = np.uint16) -> np.ndarray:
         """
         Clip the given frame to [0, 1], rescale it to [0, maxval], and convert
@@ -353,9 +379,20 @@ class Algorithms:
         agnostic.
         """
         t0 = time.time()
-        frame = self.clip(frame * maxval + 0.5, 0, maxval)
+        input_range = self._minmax(frame)
+        frame = np.clip(frame * maxval + 0.5, 0, maxval)
         frame = frame.astype(dtype)
-        self._vprint(f"{_elapsed(t0)} - quantizing from float [0, 1] to {np.dtype(dtype).name} [0, {maxval}]")
+        self._vprint(f"{_elapsed(t0)} - quantizing from [0, 1] to {np.dtype(dtype).name} [0, {maxval}]; input range = {input_range}")
+        return frame
+
+    def _quantize(self, frame: np.ndarray, maxval: int = 65535, dtype: np.dtype = np.uint16) -> np.ndarray:
+        """
+        Forced silent version of quantize().
+        """
+        verbose = self.verbose
+        self.verbose = False
+        frame = self.quantize(frame, maxval, dtype)
+        self.verbose = verbose
         return frame
 
     def _vprint(self, message: str, **kwargs):
@@ -366,20 +403,21 @@ class Algorithms:
         if self.verbose:
             minmax_str = f"[{np.min(frame):.2f}, {np.max(frame):.2f}]"
             return minmax_str
+        return ""
 
 
 def _transform_srgb_to_hsl(frame: np.ndarray) -> np.ndarray:
     rgb = frame.swapaxes(-1, 0)  # (H, W, 3) ==> (3, W, H)
-    rgb = np.clip(rgb, 0.001, 0.999)  # colorio.HSL() can't handle 0.0 and 1.0
+    rgb = np.clip(rgb, 0.001, 0.999)  # colorio.cs.HSL() can't handle 0.0 and 1.0
     with warnings.catch_warnings():
         warnings.simplefilter("ignore", category=RuntimeWarning)
-        hsl = colorio.HSL().from_srgb1(rgb)
+        hsl = colorio.cs.HSL().from_rgb1(rgb)
         hsl = np.nan_to_num(hsl)
     return hsl
 
 
 def _transform_hsl_to_srgb(frame: np.ndarray) -> np.ndarray:
-    rgb = colorio.HSL().to_srgb1(frame)
+    rgb = colorio.cs.HSL().to_rgb1(frame)
     rgb = rgb.swapaxes(-1, 0)  # (3, W, H) => (H, W, 3)
     rgb = np.ascontiguousarray(rgb)
     return rgb
